/*
    getconf.c - get configuration table from file
    Copyright (C) 1993  Uwe Ohse,  1991, 1992 Wolfram Roesler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


	For questions contact:
	e-mail: (prefered)
		Uwe_Ohse@me.maus.de
		uwe@tirka.gun.de
	snail-mail:
	    Uwe Ohse, Drosselstrasse 2, 47055 Duisburg, Germany

*/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "intern.h"
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#ifdef HAVE_FNMATCH_H
#  include <fnmatch.h>
#else
#  include "fnmatch.h"
#endif

#include "conflib.h"
#include "tinterp.h"
#include "version.h"


int cl_report_all=0;
static size_t akt_line_nr;

static int DoCmd(register char *,conf_t *);
static void zeile_auswerten(conf_t *tab, const char *stanza, char *s);
static int DupTo(conf_t *Vp, char *p);
static char *trim_in_place(char *s);
static int convert2bool(char *Ptr);
static int varnamecmp(char *p1,char *p2);
static int env_auswertung(conf_t *Konftab, long pass);

static const char *konffilename;
static const char *stanzaname;

static size_t my_private_tip_fn(const char *this, const char **next, 
	char *to, size_t maxlen);

	/* textinterpretationstabelle */
static tip_t tiptab[]={
	{ '$',	my_private_tip_fn },
	{ '$',	tip_dollar },
	{ '\\',	tip_backslash },
	{ '~',	tip_tilde },
	{ '\0', NULL }
};

static size_t my_private_tip_fn(const char *this, const char **next,
	char *to, size_t maxlen)
{
	size_t l;
	if (*this!='$')
	{
		tip_errtext="private tip_fn called without proper magic $";
		return TIP_ERROR;
	}
	switch(this[1])
	{
	case '#': /* $# -> Zeilennummer */
		{
			char buf[10];
			sprintf(buf,"%ld",(long) akt_line_nr);
			l=strlen(buf);
			*next=this+2;
			l=strlen(buf);
			if (l<maxlen)
			{
				strcpy(to,buf);
				return l;
			}
			tip_errtext="conflib tip_fn: buffer too short";
			return TIP_ERROR;
			break;
		}
	case '%': /* $% -> konfigfilename (./x.y.cnf) */
		*next=this+2;
		l=strlen(konffilename);
		if (l<maxlen)
		{
			strcpy(to,konffilename);
			return l;
		}
		tip_errtext="conflib tip_fn: buffer too short";
		return TIP_ERROR;
		break;
	case '&': /* $& -> konfigfilebasename (./x.y -> x) */
		{
			const char *p1,*p2;
			*next=this+2;
			p1=strrchr(konffilename,'/'); 
			if (!p1)
				p1=konffilename;
			else
				p1++;
			p2=strrchr(konffilename,'.');
			if (!p2)
				p2=konffilename+strlen(konffilename);
			l=(p2-p1);
			if (l<maxlen)
			{
				strncpy(to,p1,l);
				return l;
			}
			tip_errtext="conflib tip_fn: buffer too short";
			return TIP_ERROR;
		}
		break;
	case '/': /* $/ -> stanzaname or DEFAULT */
		*next=this+2;
		l=strlen(stanzaname);
		if (l<maxlen)
		{
			strcpy(to,stanzaname);
			return l;
		}
		tip_errtext="conflib tip_fn: buffer too short";
		return TIP_ERROR;
		break;
	} /* switch */
	return TIP_NOTDONE;
}

static void
fatal(char *s)
{
	/* Warum auch immer diese Routine angesprungen wird: Weitermachen
	 * hat keinen Zweck mehr, da ein Programmierfehler vorliegt.
	 */
	fprintf(stderr,"conflib: %s (aborting, internal program error)\n",
		s);
	exit(1);
}

static 
conf_t *unalias(conf_t *tab, conf_t *entry)
{
	int derefed=0;
	char *name;
	while (entry->typ == CL_ALIAS)
	{
		conf_t *akt;
		if (cl_report_all)
		{
			if (!derefed)
			{
				fprintf(stderr,"Warning: Alias %s used, ",
					entry->varname);
			}
		}
		akt=tab;
		name=(char *) entry->secdata;
		if (!name)
			fatal("alias has no name!");
		while (akt->varname)
		{
			if (!varnamecmp(name,akt->varname))
				break;
			akt++;
		}
		if (!akt->varname)
			fatal("alias not dereferenced");
		derefed++;
		entry=akt;
	}
	if (cl_report_all && derefed)
	{
		fprintf(stderr,"use %s instead\n",
			entry->varname);
	}
	return entry;
}
	
static int varnamecmp(char *p1,char *p2)
{
	while (*p1)
	{
		if (*p1=='-' || *p1=='_')
		{
			if (*p2!='-' && *p2!='_')
				return (*p1 - *p2);
		}
		else if (tolower(*p1)!=tolower(*p2))
			return tolower(*p1)-tolower(*p2);
		p1++;
		p2++;
	}
	return *p1 - *p2;
}

static char *trim_in_place(char *s)
{
	char *p;
	p=s;
	while (*p==' ' || *p=='\t')
		p++;
	s=p+strlen(p)-1;
	while (s>=p && (*s==' ' || *s == '\t' || *s=='\n' || *s=='\r'))
	{
		*s='\0';
		s--;
	}
	return (p);
}


int cl_getconf(const char *FName,conf_t *KonfTab)
{
	return cl_getconfstanza(FName,KonfTab,(char*)0,0);
}

static int convert2bool(char *Ptr)
{
	/* no nice code, indeed */
	switch(*Ptr)
	{
		case 't':	/* true */
		case 'T':
		case 'Y':	/* yes */
		case 'y':  
		case 'J':	/* ja */
		case 'j':  
		case 'w':   /* wahr */
		case 'W': 
			return 1;
			break;
		case 'F':	/* False, Falsch */
		case 'f':
		case 'N':	/* no, nein */
		case 'n':  
			return 0;
			break;
		default:
			/* everything else is assumed numeric,
				0 is false, !=0 means true
			*/
			return (int) (atoi(Ptr) != 0);
			break;
	}
}

static int in_stanza=0;

static void 
zeile_auswerten(conf_t *tab, const char *stanza, char *s)
{
	static int in_comment=0;
	char *p;

	if (!s)
	{
		in_comment=0;
		in_stanza=0;
		return;
	}

	p = trim_in_place(s);
	if (*p=='\0' || *p == '#')
	{
		if (*p=='#' && p[1]=='#')
			in_comment = !in_comment;
		return;
	}
	if (!in_comment)
	{
		if (in_stanza)
		{
			if (!strchr(p,'=') && strchr(p,':'))
				in_stanza=0;
		}
		if (in_stanza)
		{
			if (!strcmp(p,"end"))
			{
				in_stanza=0;
				if (cl_report_all)
				{
					fprintf(stderr,
						"'end' statement in configfiles is obsolete\n");
				}
			}
			else
				DoCmd(s,tab);
		}
		else
		{
			/* vor einer Stanza */
			in_stanza=0;
			if (!strchr(p,'=') && p[strlen(p)-1]==':')	/* Label? */
			{
				char *p2;
				for	(p2=strtok(p," \t:"); p2; p2=strtok(NULL," \t:"))
				{
					if (!fnmatch((char *)stanza,p2,0))
					{
						in_stanza=1;
						break;
					}
				}
			}
		}
	}
}

static FILE *conf_f=NULL;

void 
cl_closeconf(void)
{
	if (conf_f!=NULL)
	{
		fclose(conf_f);
		conf_f=NULL;
	}
}

int 
cl_getconfstanza(const char *FName,conf_t *KonfTab,const char *Stanza,
	int keep_open)
{
	char St[1024];
	int Ret=0;

	zeile_auswerten(NULL,NULL,NULL);
	
	if (Stanza==NULL)
		in_stanza = 1;

	konffilename=FName;
	stanzaname=Stanza ? Stanza : "*";
	if (!conf_f)
	{
		tip_init();
		conf_f = fopen(FName,"r");
	}
	else
		rewind(conf_f);
	if (conf_f!=NULL)
	{
		char *p;
		/* Env-Variablen, Preauswertung */
		Ret=env_auswertung(KonfTab,CL_PREENV);
		akt_line_nr=0;

		while(fgets(St,(int)sizeof(St),conf_f))
		{
			akt_line_nr++;
			p=strpbrk(St,"\r\n");
			if (p)
				*p--='\0';
			else
				p=strlen(p)+p-1;
			while (*p=='\\')
			{
				char *p2;
				int noch_frei;
				/* Zeile ist noch nicht zu Ende */
				*p=' ';
				noch_frei=(int) (sizeof(St) - (p-St) -1);
				if (noch_frei<=0 || !fgets(p+1,noch_frei,conf_f))
					break;
				p2=strpbrk(p,"\r\n");
				if (p2)
				{	
					*p2='\0';
					p=p2-1;
				}
				else
					p=p+strlen(p)-1;
			}
			zeile_auswerten(KonfTab, Stanza, St);
		}
		if (!keep_open)
		{
			fclose(conf_f);
			conf_f=NULL;
		}
	}
	else
		Ret=1;
	if (!Ret)
		Ret=env_auswertung(KonfTab,CL_POSTENV);
	tip_deinit();

	return Ret;
}

static int env_auswertung(conf_t *KonfTab, long pass)
{
	conf_t *Vp;
	char *Ptr;
	int Ret=0;
	
	/* Env-Variablen, Postauswertung */
	for(Vp=KonfTab; Vp->varname; Vp++)
	{
		if (!Vp->secdata)
			continue;
		if (!(Vp->flags & pass))
			continue;
		Ptr = getenv((char *)Vp->secdata);
		if (Ptr)
		{
			switch(Vp->typ)
			{
				case CL_STRING:
				case CL_ISTRING:
					if ((Vp->flags & CL_MALLOCED) 
						&& (char *)(Vp->adr)!=(char *)NULL)
					{
						free(*(char **)Vp->adr);
						Vp->flags-=CL_MALLOCED;
					}
					if (DupTo(Vp, Ptr))
					{
						Ret=1;
					}
					break;
				case CL_NUM:
					*(long*)Vp->adr = atol(Ptr);
					break;
				case CL_NUMSHORT:
					*(short*)Vp->adr = (short) atoi(Ptr);
					break;
				case CL_BOOLEAN:
					*(int *)Vp->adr=convert2bool(Ptr);
					break;
				case CL_ARRAY:
				case CL_IARRAY:
				case CL_ENUM:
				case CL_ALIAS: /* keep gcc happy */
				case CL_DUMMY:
				case CL_LIST:
					break;
			}
		}
	}
	return Ret;
}
	
static int DupTo(conf_t *Vp, char *p)
{
	size_t l=strlen(p);
	char *ad;
	if (Vp->flags & CL_MALLOCED)
	{
		Vp->flags -=CL_MALLOCED;
		free(*(char **)Vp->adr);
	}
	ad=malloc(l+1);
	if (!ad)
		return 1;
	Vp->flags|=CL_MALLOCED;
	*(char **)Vp->adr=ad;
	strcpy((char *)ad,p);
	return 0;
}

static int DoCmd(register char *Ptr,conf_t *KonfTab)
{
	register char *p;
	register conf_t *Vp;
	static int report_unknown=1;
	conf_enum_t *enump;
	int found=0;

	if (!strcmp(Ptr,"GetKonfReportUnknown"))
	{
		if (cl_report_all)
		{
			fprintf(stderr,"Warning: GetKonfReportUnknown used,"
				" ignored for now\n");
			return 0;
		}
		report_unknown=1;
		return 0;
	}
	if (!strcmp(Ptr,"GetKonfDontReportUnknown"))
	{
		if (cl_report_all)
		{
			fprintf(stderr,"Warning: GetKonfDontReportUnknown used,"
				" ignored for now\n");
			return 0;
		}
		report_unknown=0;
		return 0;
	}
	
	/* '=' finden */
	p = strchr(Ptr,'=');
	if (p==NULL)
	{
		if (report_unknown)
			fprintf(stderr,"getkonf: unknown cmd '%s'\n",Ptr);
		return 0;
	}
	if (*Ptr=='$')
	{
		/* Dann diese Variable dem Environment zufuegen */
		char *s;
		s=malloc(strlen(Ptr+1));
		if (!s)
		{
			fprintf(stderr,"getkonf: out of memory\n");
			return(1);
		}
		strcpy(s,Ptr+1);
		if (putenv(s))
		{
			fprintf(stderr,"getkonf: out of memory\n");
			return(1);
		}
		return 0;
	}
	*p++ = '\0';

	Ptr=trim_in_place(Ptr);
	p=trim_in_place(p);
	/*
		p 	zeigt nun auf den Wert
		Ptr	auf den Variablennamen 
	*/
	for(Vp=KonfTab;Vp->varname;Vp++)
	{
		found=varnamecmp(Ptr,Vp->varname)==0;
		if (found)
		{
			Vp=unalias(KonfTab,Vp);
			switch(Vp->typ)
			{
			case CL_ENUM:
				if (!Vp->secdata)
				{
					fprintf(stderr,
						"conflib warning: CL_ENUM nullpointer dereference!\n"
						);
					return 1;
				}
				for (enump=(conf_enum_t *) Vp->secdata; enump->magic; enump++)
				{
					if (!strcasecmp(enump->magic,p))
					{
						*(long*)Vp->adr = enump->value;
						return 0;
					}
				}
				fprintf(stderr,
					"conflib warning: unknown value %s\n",Ptr
					);
				*(long*)Vp->adr = (long) 0;
				return 1;
				break;
			case CL_ISTRING:
				{
					char buf[4096];
					if (tip_interpret(tiptab, p, buf, sizeof(buf))==TIP_ERROR)
						return 1;
					return DupTo(Vp,buf);
				}
			case CL_STRING:	/* Stringvariable */
				return DupTo(Vp,p);
			case CL_NUM:
				*(long*)Vp->adr = atol(p);
				return 0;
			case CL_NUMSHORT:
				*(short*)Vp->adr = atoi(p);
				return 0;
			case CL_BOOLEAN:
				*(int *)Vp->adr=convert2bool(p);
				return 0;
			case CL_ARRAY:	/* Stringarray */
			case CL_IARRAY:	/* mit Env-Expand */
				{
					register int Cnt;
					register char **Ap;
					for(Ap=(char**)Vp->adr,Cnt=0;*Ap && Cnt<CL_MAXARRAY;Ap++,Cnt++);
					if (Cnt<CL_MAXARRAY)
					{
						if (Vp->typ==CL_ARRAY)
						{
							*Ap = malloc(strlen(p)+1);
							if (*Ap)
								strcpy(*Ap,p);
						}
						else
						{
							char buf[4096];
							if (tip_interpret(tiptab, p, buf, sizeof(buf))
									==TIP_ERROR)
								return 1;
							*Ap=strdup(buf);
							if (!*Ap)
							{
								fprintf(stderr,"conflib: out of memory\n");
								return 1;
							}
						}
						Ap[1]=0;
					}
					return 0;
				}
				break;
			case CL_ALIAS:
				break;
			case CL_DUMMY:
				if (cl_report_all)
				{
					fprintf(stderr,"Warning: dummy %s used, please remove\n",
						Vp->varname);
				}
				break;
			case CL_LIST:
				{
					char *isep;
					char *this;
					char *psep;
					conf_list_t *last;
					conf_list_t *new;
					last=*(conf_list_t **)Vp->adr;
					while (last && last->next)
						last=last->next;

					isep=(char *)Vp->secdata;
					if (!isep)
						isep=":\t ";

					this=p;
					while (this && *this && strchr(isep,*this))
						this++;
					while (this && *this)
					{
						psep=strpbrk(this,isep);
						if (psep)
							*psep++='\0';
						new=malloc(sizeof(conf_list_t));
						if (new)
							new->inhalt=strdup(this);
						if (!new || !new->inhalt)
						{
							fprintf(stderr,"getkonf: out of memory\n");
							return(1);
						}
						new->next=NULL;
						if (last)
							last->next=new;
						else
							*(conf_list_t **)Vp->adr=new;
						last=new;
						this=psep;
						while (this && *this && strchr(isep,*this))
							this++;
						Vp->flags|=CL_MALLOCED;
					} /* while sep */
					return 0;
				}
				break;
			} /* switch */
			return 0;
		}
	} /* for */
	if (report_unknown)
		fprintf(stderr,"getkonf: unknown cmd '%s'\n",Ptr);
	return 0;
}

int cl_isinarray(const char **Array,const char *St,int insensitive)
{
	if (!Array)
		return 0;
	while(*Array)
	{
		if (insensitive)
		{
			if (!strcasecmp(St,*Array))
				return 0;
		}
		else
		{
			if (!strcmp(St,*Array))
				return 0;
		}
		Array++;
	}
	return 1;
}

int cl_isinwarray(const char **Array,const char *St)
{
	while(*Array)
	{
		if (!fnmatch(St,*Array,0))
			return 0;
		Array++;
	}
	return 1;
}

int cl_getversion(void)
{
	return INTEGERVERSION;
}

void cl_clearconf(conf_t *tab)
{
	conf_t *Vp;

	for(Vp=tab;Vp->varname;Vp++)
	{
		switch(Vp->typ)
		{
		case CL_ENUM:
			*(long*)Vp->adr = (long) 0;
			break;
		case CL_ISTRING:
		case CL_STRING:	/* Stringvariable */
			if (*(char **)Vp->adr && Vp->flags & CL_MALLOCED)
			{
				Vp->flags-=CL_MALLOCED;
				free(*(char **)Vp->adr);
				*(char **)Vp->adr=NULL;
			}
			break;
		case CL_NUM:
			*(long*)Vp->adr = (long) 0;
			break;
		case CL_NUMSHORT:
			*(short*)Vp->adr = (short) 0;
			break;
		case CL_BOOLEAN:
			*(int *)Vp->adr= (int) 0;
			break;
		case CL_ARRAY:	/* Stringarray */
		case CL_IARRAY:	/* mit Env-Expand */
			{
				char **atab;
				int i;

				i=0;
				atab=(char**)Vp->adr;
				while (*atab && i<CL_MAXARRAY)
				{
					free(*atab);
					*atab=NULL;
					atab++;
					i++;
				}
			}
			break;
		case CL_ALIAS:
		case CL_DUMMY:
			break;
		case CL_LIST:
			{
				conf_list_t *akt;
				conf_list_t *next;
				if (!Vp->flags & CL_MALLOCED)
					break;
				akt=*(conf_list_t **)Vp->adr;
				while (akt)
				{
					if (akt->inhalt)
						free(akt->inhalt);
					next=akt->next;
					free(akt);
					akt=next;
				}
				*(conf_list_t **)Vp->adr=NULL;
				Vp->flags-=CL_MALLOCED;
			}
			break;
		} /* switch */
	} /* for */
}
