/* lib/smtp.c: Routinen fr smtp. */

#include "ownstd.h"

#ifdef HAVE_SYS_STAT_H
#  include <sys/stat.h>
#endif
#include <errno.h>
#include <ctype.h>
#include <conflib.h>
#include <stdarg.h>

#include "mausconlib.h"
#include "metamsg.h"
#include "rfc.h"
#include "smtp.h"

#ifndef HAVE_TCP
#undef FEATURE_SMTP
#endif

#define CRLF "\r\n"


#if defined(FEATURE_SMTP)
static char smtprepbuffer[1024];
static char *format_reply(void);
static int retcode_and_log(int r);
#endif

#if defined(FEATURE_SMTP)
static char *
format_reply(void)
{
	char *p=smtprepbuffer;
	char *p2;
	while (isspace(*p) && isdigit(*p))
		p++;
	p2=strlen(p)+p-1;
	while (*p2=='\r' || *p2=='\n')
		*p2--='\0';
	return p;
}

void smtp_close(FILE *f)
{
	fprintf (f, "QUIT" CRLF);
	(void) smtp_get_reply(f,NULL);
	fclose(f);
}

FILE *
smtp_open(char *host)
{
	int ret;
	FILE *f;

	f = fopen_service ("smtp", host);
	if (!f)
	{
		/* Die Fehlermeldung wurde von open_service schon geschrieben */
		return NULL;
	}

	setvbuf (f, NULL, _IOLBF, 0);

	ret = smtp_get_reply (f,NULL);	/* Connection establishment reply */
	if (ret != 220)
	{
		/* Tja, keine Verbindung ... */
		log(LOG_ERR,"Connection not established");
		fclose (f);
		return NULL;
	}
	/*
	 * Schade, sendmail mag die naechste Zeile nicht.
	 * fprintf (f, "HELO %s%s" CRLF, opts.pollsite, 
	 *    opts.mausdomain);
	 */
	fprintf (f, "HELO %s" CRLF, opts.mysitedomain);
	ret = smtp_get_reply (f,NULL);	/* HELO reply */
	if (ret >= 300)
	{
		log (LOG_ERR,"HELO failed");
		fclose (f);
		return NULL;
	}
	return f;
}

int
smtp_get_reply (FILE *f, char **s)
{
	fd_set fds;
	struct timeval timeout;
	int cont = 1;

	FD_ZERO (&fds);
	FD_SET (fileno (f), &fds);

	/* Initialize the timeout data structure. */
	/* RFC 1047 sagt, da man zumindest nach DATA *lange* warten sollte,
	 * weil die Nachrichtenverarbeitung durchaus eine Weile dauern kann.
	 * Zitat:
	 *   the increasing round-trip times on the Internet, and that some
	 *   processing after the final dot is required, the timeout for reply to
	 *   the final dot should probably be at least 5 minutes and a timeout of
	 *   10 minutes would not be unreasonable.
	 */

	timeout.tv_sec = 10 * 60;			/* seconds */
	timeout.tv_usec = 0;
	if (s)
		*s=smtprepbuffer;

	/* `select' returns 0 if timeout, 1 if input available, -1 if error. */
	if (select (FD_SETSIZE, NULL, &fds, NULL, &timeout)!=1)
	{
		*smtprepbuffer='\0';
		return 421;				/* Timeout signalisieren */
	}

	while (cont)
	{
		fgets (smtprepbuffer, sizeof (smtprepbuffer), f);
		if (smtprepbuffer[3] != '-')
			cont = 0;
	}
	return (int) strtol (smtprepbuffer, NULL, 10);
}

static int 
retcode_and_log(int r)
{
	switch(r)
	{
	case 421: /* Service not available, closing transmission channel */
		/* auch timeout! */
		log(LOG_ERR,"Verbindung verloren");
		return 2;
	case 500: /* Syntax error, command unrecognized */
	case 501: /* Syntax error in parameters or arguments */
	case 502: /* Command not implemented */
	case 503: /* Bad sequence of commands */
	case 504: /* Command parameter not implemented */
		/* Weitermachen sinnlos. */
		log(LOG_ERR,"%s",format_reply());
		return 2;
	case 551: /* User not local, please try <forward-path> */
		msglog("%s",format_reply());
		return 1; /* Auf einem anderen Weg nochmal versuchen */
	case 552: /* exceeded storage allocation */
	case 553: /* mailbox name not allowed */
	case 554: /* Transaction failed */
	case 450: /* mailbox unavailable / busy */
	case 451: /* local error in processing */
	case 453: /* insufficient system storage */
		msglog("%s",format_reply());
		return 1;
	default: /* Was macht man mit unbekannten Meldungen? */
		if (r>=500)
		{
			log(LOG_ERR,"%s",format_reply());
			return 2;
		}
		log(LOG_WARNING,"%s (unknown reply)",format_reply());
		msglog("%s",format_reply());
		return 1;
	}
}
#endif /* FEATURE_SMTP && HAVE_TCP */

#define REPLY(datei) ((no_reply) ? 200 : smtp_get_reply(datei,NULL))
int smtp_mail(FILE *f,metamsg_t *meta,int no_reply)
{
	int ret;
	char zwi[2048];
	/* SMTP-Envelope */
	sprintf (zwi, "MAIL FROM:<%s>" CRLF, meta->env_from);
	fprintf (f, "%s", zwi);
#ifdef FEATURE_SMTP
	ret = REPLY(f);	/* MAIL reply */
	if (ret > 299)
	{
		return retcode_and_log(ret);
	}
#endif
	sprintf (zwi, "RCPT TO:<%s>" CRLF, meta->env_to);
	fprintf (f, "%s", zwi);
#ifdef FEATURE_SMTP
	ret = REPLY(f);	/* MAIL reply */
	if (ret>299)
	{
		return retcode_and_log(ret);
	}
#endif

	fprintf (f, "DATA" CRLF);
#ifdef FEATURE_SMTP
	ret = REPLY(f); 
	/* Bei irgendeiner Form von Fehler die Nachricht
	 * in einen Mailbatch schreiben. Es gibt nichts, was
	 * sonst getan werden kann.
	 */
	if (ret>399)
	{
		return retcode_and_log(ret);
	}
#endif
	/* Nachricht schreibeni, Zeilentrenner ist CRLF, Punkte escapen */
	rfc_print_message(meta,f,CRLF,1);
	fprintf (f, "." CRLF);
#ifdef FEATURE_SMTP
	ret = REPLY(f);
	if (ret>299)
	{
		return retcode_and_log(ret);
	}
#endif
	return 0;
}
#undef REPLY

